import redis
from typing import Union
from fastapi import FastAPI, Request
r = redis.Redis.from_url(url="redis://:1234qwer666@150.158.144.155:6390/0", decode_responses=True)
app = FastAPI()


def test():
    myredis = redis.Redis.from_url(url="redis://:1234qwer666@150.158.144.155:6390/0", decode_responses=True)
    # myredis = redis.Redis(host='150.158.144.155', port=6390, decode_responses=True,
    #                 username="", password="1234qwer666", db=0)
    # ipdb.set_trace()
    myredis.set('foo', 'bar')
    myredis.set('xxxlock', 'client_id', nx=True, ex=600)
    print(myredis.get('foo'))
    print(myredis.get('foo'))
    print(myredis.get('foo'))


@app.get("/set/{key}/{value}", tags=["redis"])
def get_customer(key: str, value: str):
    r.set(key, value)
    return {"redis": "set", key: value}


@app.get("/get/{key}", tags=["redis"])
def get_customer(key: str):
    value = r.get(key)
    return {"redis": "get",  key: value}


@app.get("/ratelimit/{key}", tags=["redis"])
def ratelimit(key: str, ):
    r.set(key, 0, ex=10, nx=True)
    v = r.incr(key)
    ttl = r.ttl(key)
    return {"redis": "setnx", "key": key, "counter": v, "ttl": ttl}



### todo:
# 将redis的其他数据结构实现为 fastapi 的一个endpoint(接口)


#  运行程序:
#  uvicorn.exe redisdemo:app



# 此版本用于研究 python 实现 zen expression 的解析, 进而实现python自定义函数和zen表达式的混合调用.
# 此模块处于研究状态.
# func_engine 此模块要改名为 udf_engine 或者是 func_engine.
# 此模块完成函数解析
## 完成 foo(bar(2  , zoo(3,6),'a'), bas())

# https://tree-sitter.github.io/tree-sitter/using-parsers/1-getting-started.html
# https://stackoverflow.com/questions/67656240/parsing-a-function-call-using-python-regex
# https://regex101.com/r/Umsdg0/1
import asyncio
import logging
from pprint import pprint, pformat
from dataclasses import dataclass, asdict, field
from pathlib import Path
import zen
from .udf_manager import udf_manager
from .func_parser import zen_custom_expr_parse, FuncItem
logger = logging.getLogger(__name__)


args_expr_eval = lambda x, input: zen.evaluate_expression(x, input)


async def ast_exec(expr_ast, args_input, context={}):
    """
        实现自定义函数的嵌套调用.
    """
    func_value_context = {}  # ast 求值时, 函数值暂存在此字典, 用于嵌套函数传参.
    # logger.debug(f"{pformat(ast)}")
    for func in expr_ast:  # 执行一个嵌套函数表达式.
        ### 下列代码需要封装为一个执行引擎.
        logger.debug(f"current func_value_context: {func_value_context}")
        if func["ns"] == "udf":
            # logger.warning(f"udf expression: {func}")
            func_name = func["name"]
            args = FuncItem.args_eval(func["args"], func_value_context, args_input, args_expr_eval)
            kwargs = {
                **context,
                "func_id": func["id"],
            }
            result = await udf_manager(func["name"], *args, **kwargs)
            logger.warning(f"{func_name}({args}) ->: {result}")
            func_value_context[func["id"]] = result
        elif func["ns"] == "":
            func_name = func["name"]
            args = func["args"]
            # 需要判断参数是嵌套函数的情况.
            args = [func_value_context.get(func["id"]) if t == "func_value" else str(arg) for arg, t in args]
            # 要判断 args 中是否还嵌套函数.
            func_call_args = ",".join(args)
            zen_expr = f"{func_name}({func_call_args})"
            result = zen.evaluate_expression(zen_expr, args_input)
            logger.warning(f"{zen_expr} ->: {result}")
            # 默认使用 zen 表达式.
            func_value_context[func["id"]] = result
        else:
            raise RuntimeError(f'自定义函数{func["name"]}表达式不支持')

    return result


if __name__ == "__main__":
    ### 在项目根目录运行 python -m src.custom.func_engine_v3 即可测试此方法.

    # "udf:foo(udf:bar(2  , rand(100),'a'), bas())"  "udf:foo(udf:bar(rand(100), udf:zoo(3,6),'a'), udf:bas())"
    func_call_s = "udf:foo(myvar,udf:bar(rand(100), udf:zoo('fccdjny',6, 3.14),'a'), udf:bas())"
    # func_call_s = "foo(bar( 2,'a'), bas())"
    # func_call_s = "rand(100)"
    # func_call_s = "map(['a', 'b', 'c'], # + '!')"  # 增加对 ['a', 'b', 'c'] 数组类型的解析.         # 已经实现支持
    # func_call_s = "map(['a', lower('b'), 'c'], # + '!')"  # 增加对 ['a', 'b', 'c'] 数组类型的解析.  # 已经实现支持
    # func_call_s = '{customer: { firstName: "John", lastName: "Doe"}}'  # 增加对 json5 objects(结构)类型的解析. 这个表达式不是函数调用, 不会被解析.
    # func_call_s = "map([{id: 1, name: 'John'}, {id: 2, name: 'Jane'}], #.id)"  # 增加对数组和objects类型的解析.

    # # 这些语法和数组语法解析冲突, 需要在参数解析中来做分区.
    # func_call_s = "map([0..3], #)"   # 暂不支持
    # func_call_s = "map((0 .  .   3], #)"   # 暂不支持, 报错.

    ### ************************************************************************************************************
    ## 注意, zen-engine 又加入了很多新的语法, 从长远来说, 这些语法都得兼容, 否则就无法实现 python udf 和 zen 表达式函数的混合调用.
    ## 把用 python 实现 zen expression 的解析(得到 zen 表达式的抽象语法树ast)作为一个长期的目标去支持吧.
    ### ************************************************************************************************************

    expr_ast = zen_custom_expr_parse(func_call_s)
    print("expr    :", func_call_s)
    print("expr_ast:", expr_ast)
    pprint(expr_ast)
    import json
    rrr = json.dumps(expr_ast)
    print(rrr)

    res = asyncio.run(ast_exec(expr_ast, {"input": 7, "myvar": 15}))
    print("res:", res)


import logging
from pathlib import Path
from pprint import pprint, pformat
import zen
import asyncio
import json
import zen

from tools import contexts
from udfs.udf_manager import udf_manager
from cachetools import LRUCache
from cachetools import cached
from config import CACHE_SIZE
from db.models.rule_meta import BrdeRule
from db.postgresql.database import SessionLocal
from db.postgresql.crud import read_one
from db.models.rule_meta import BrdeUser, BrdeProj, BrdeRule, BrdeList, ListData

lru_cache = LRUCache(maxsize=CACHE_SIZE)
cache_dict = {}

logger = logging.getLogger(__name__)


class RuleManager:

    async def create_rule_manager(self, data_dict,  rule_id=None, db=None, **kwargs):
        await self._init(data_dict, rule_id, db, **kwargs)
        return self 

    def __init__(self):
        self.meta = {}
        self.cache_set = {}
        self.engine = None
        self.decision = None
    
    async def _init(self, data_dict, rule_id, db, **kwargs):
        engine = self.create_engine()
        self.engine = engine
        decision, _meta = await self._init_decsion(data_dict, rule_id, db, **kwargs)
        self.decision = decision
        self.meta = _meta
        if rule_id:
            await self._init_cache(rule_id)
        return

    async def custom_handler(self, request, **kwargs):
        funcs = request.node["config"].get("inputs", [])
        trans_func = lambda x: zen.evaluate_expression(x, request.input)
        res = {}
        bar = []
        for item in funcs:
            funcmeta = item["funcmeta"]
            funcname = funcmeta["name"]
            func_variables = {k: v for k, v in item["arg_exprs"].items()}
            vars2value = {k: trans_func(v) for k, v in func_variables.items()}
            env = {
                "node_id": request.node["id"],  ## 隔离 graph 中的节点
                "func_id": item["id"],  ## 隔离每一个计算逻辑.
                "meta": request.node["config"].get("meta", {})
            }
            env.update(vars2value)
            bar.append((item["key"], funcname, env))
        results = await asyncio.gather(*[udf_manager.call_udf(_[1], *_[2], **_[2], **kwargs) for _ in bar])
        for key, result in zip([_[0] for _ in bar], results):
            res[key] = result
        res.update({k: v for k, v in request.input.items() if k != "$nodes"})
        return {
            "output": res
        }

    def create_engine(self):
        engine = zen.ZenEngine({"customHandler": self.custom_handler})
        return engine

    def add_graph_meta(self, rule_graph):
        input_node_name = [i.get("name") for i in rule_graph['content']["nodes"] if i.get("type") == "inputNode"][0]
        for node in rule_graph['content']["nodes"]:
            if node.get("type") == "customNode":
                content = node.get("content", {})
                config  = content.get("config", {})
                meta = config.get("meta", {})
                meta["inputNode_name"] = input_node_name
                node["content"]["config"]["meta"] = meta
            else:
                meta = {}
        return json.dumps(rule_graph['content']), meta
    
    async def get_graph_data_from_db(self, rule_id, db, **kwargs):
        rule_data = await read_one(db, BrdeRule, {"id": rule_id})
        rule_dict = rule_data.as_dict(all_columns=True)
        jdm_graph = json.loads(rule_dict['rule_graph'])
        jdm_str, _meta = self.add_graph_meta(jdm_graph)
        return jdm_str, _meta 
    
    def get_graph_data_from_content(self, data_dict):
        jdm_str = json.dumps(data_dict['content'])
        return jdm_str
    
    async def _init_decsion(self, data_dict, rule_id, db, **kwargs):
        _meta = {}
        if rule_id:
            jdm_str, _meta = await self.get_graph_data_from_db(rule_id, db, **kwargs)
        else:
            jdm_str = self.get_graph_data_from_content(data_dict)
        decision = self.engine.create_decision(jdm_str) 
        return decision, _meta

    async def _init_cache(self, rule_id):
        cache_dict = {
            'meta': self.meta,
            'engine': self.engine,
            'decision': self.decision
        }

        self.cache_set[rule_id] = cache_dict
        return
        
    async def rule_exec(self, data_dict, rule_id, trace, db, **kwargs):
        if rule_id in self.cache_set:
            rule_cache = self.cache_set[rule_id]
            decision = rule_cache['decision']
        else:
            rule_manager = await self.create_rule_manager(data_dict, rule_id, db, **kwargs)
            decision = rule_manager.decision
        logger.debug('rule exec info: rule_id:{}, rule_manager:{} ,decision:{}'.format(rule_id, self, decision))
        result = await decision.async_evaluate(data_dict['context'], {"trace": trace})
        return result

    def update_rule_manager(self, rule_id):
        cache_dict = {
            'meta': self.meta,
            'engine': self.engine,
            'decision': self.decision
        }
        self.cache_set[rule_id] = cache_dict

    def delete_rule_manager(self, rule_id):
        if rule_id in self.cache_set:
            del self.cache_set[rule_id]

class zenRule:
    def __init__(self, options_dict=None):
        self.options_cache = {}
        self.options_dict = options_dict
        self.engine = zen.ZenEngine(self.options_dict)

    def create_decision(self, content):
        return self.engine.create_decision(content)

    def get_decision(self, key):
        zendecision = self.options_cache.get(key, None)
        if not zendecision:
            zendecision = self.engine.get_decision(key)
            self.options_cache[key] = zendecision
            return zendecision
        else:
            return zendecision

    def evaluate(self, key, ctx, options_dict=None):
        return self.engine.evaluate(key, ctx, options_dict)

    def async_evaluate(self, key, ctx, options_dict=None):
        return self.engine.async_evaluate(key, ctx, options_dict)

rule_manager = RuleManager()


import asyncio
import logging
from pathlib import Path
from pprint import pprint, pformat
import json
import contextvars
import pytest
import zen
import aiohttp
import redis
logger = logging.getLogger(__name__)
context_http_client = contextvars.ContextVar('http_client', default=None)

rule_basedir = Path(__file__).parent / "test-data-brde"

class RuleAssert:
    def __init__(self, rule_content, **kwargs) -> None:
        # self.meta = meta
        self.user_name = kwargs.get("user_name")
        self.inputNode_name = kwargs.get("inputNode_name")
        con, _meta = self.open_graph_with_meta(rule_content)
        self.rule_content = con
        self.meta = _meta
        self.engine = zen.ZenEngine({"customHandler": self.custom_async_handler})
        self.decision = self.engine.create_decision(con)  # 耗时操作, 所以需要复用 decision 对象实例.
        # decision.__setattr__("meta", meta)   # AttributeError: 'builtins.ZenDecision' object has no attribute 'meta'

    def update_decision(self, rule_content):
        con, _meta = self.open_graph_with_meta(rule_content)
        self.rule_content = con
        self.meta = _meta
        self.decision = self.engine.create_decision(con)

    async def async_evaluate(self, input, context: dict={}):
        # input.update({"__meta__": self.meta})
        # result = await self.decision.async_evaluate(input)  # 也许这种字典本地改动性能会好一些.
        self.meta.update(context)
        result = await self.decision.async_evaluate({**input, **{"__meta__": self.meta}})
        return result

    async def custom_async_handler(self, request):
        """
            一旦有异步自定义函数，需要使用异步方法去求表达式的值

            为了便于拍错, 最后自定义函数最外层包裹一层异常捕获, 否则这些异常会再 rust 中捕获, 无法去控制出错信息的渲染.
            比如打印异常栈方便定位错误.
        """
        http_client_session  = context_http_client.get()
        logger.warning(f"http_client_session:{http_client_session}")
        async with http_client_session.get('http://httpbin.org/get') as resp:
                    assert resp.status == 200
                    res = await resp.json()
                    assert res["url"] == 'http://httpbin.org/get'
                    logger.warning(f"httpbin get:{res}")
        logger.info(f"request.input:\n{pformat(request.input)}")
        logger.info(f"request.node:\n{pformat(request.node)}")
        request_node = request.input.get("$nodes", {}).get("Request", {})
        # pprint(request_node)
        prop1 = request.get_field("prop1")
        prop1_raw = request.get_field_raw("prop1")
        await asyncio.sleep(0)
        return {
            "output": {
                        "prop1_raw": prop1_raw,
                        "prop1": prop1,
                        "request.input": request.input,
                        "request.node": request.node,
                    }
        }

    def open_graph_with_meta(self, rule_content):
        # 此规则图只包含一个自定义节点.   请求 --> 自定义节点 --> 输出
        # 在自定义节点的 config meta 中添加输入节点的 name, 用于后续自定义函数通过 request.input 定位输入节点中
        # 传入的相关实例, 比如数据库的连接池, http的连接池, 业务参数, trace_id等
        rule_graph = json.loads(rule_content)
        name = [i.get("name") for i in rule_graph["nodes"] if i.get("type") == "inputNode"][0]
        for node in rule_graph["nodes"]:
            if node.get("type") == "customNode":
                content = node.get("content", {})
                config  = content.get("config", {})
                meta = config.get("meta", {})
                meta["inputNode_name"] = name
                # print("meta:", meta)
                node["content"]["config"]["meta"] = meta
            # print("node:", node)
        return json.dumps(rule_graph), meta


@pytest.mark.asyncio
async def test_custmer_node_cache_decision_contain_context_input():
    """
        request --> expression --> customer --> response
        customer 通过自定义函数可以获取上一个节点的输出和输入节点内容.
        通过此逻辑，可以在自定义函数中传递一些单例实例, 比如数据库的连接池, http的连接池, 业务参数, trace_id等.
    """
    # 方式二, 把meta信息写在 decision 对象中.
    # con, meta = open_graph_with_meta()
    # meta["redispool"] = "redispool"
    # meta["httpclient"] = "httpclient"
    # rule_basedir / "custom2.json"
    rule_path = rule_basedir / "custom2.json"
    with open(rule_path) as f:
        con = f.read()

    session = aiohttp.ClientSession()
    logger.warning(f"aiohttp session:{session}")
    context_http_client.set(session)
    # context_http_client.set(session) 必须要在 zen.ZenEngine({"customHandler": self.custom_async_handler}) 之前.
    # 否则 context_http_client 无法取到 session.
    rule_assert = RuleAssert(con)   # 缓存外层我们的规则包装对象.
    inputs = {"a": 10, "b": "cjn"}
    ### -------------------------------
    res = await rule_assert.async_evaluate(inputs, {"user": "brde@geetest.com"})
    assert res["result"]["prop1_raw"] == '{{ a + 10 }}'
    # 如果修改了规则图中的自定义节点的表达式, 需要重新修改这里的断言条件.
    assert res["result"]["prop1"]     == res["result"]["request.input"]['a'] + 10

    # 规则图中添加的元信息和自定义节点读取的元信息是否一致
    custome_request_node = res["result"]["request.node"]
    meta = rule_assert.meta
    assert meta["inputNode_name"] == custome_request_node["config"]["meta"]["inputNode_name"], "规则图中添加的元信息和自定义节点读取的元信息不一致"

    all_execd_nodes =  res["result"]["request.input"]["$nodes"]
    request_nodes_inuput = all_execd_nodes[meta["inputNode_name"]]
    del request_nodes_inuput["__meta__"]
    assert request_nodes_inuput == inputs, "输入和自定义节点中 request.input $nodes 中的输入不同"
 
    # 一般可以直接在测试函数的末尾 assert False 让测试失败, 然后使用 -l 选项输出测试失败时的局部变量的值. 方便排查问题.
    # result = res["result"]
    # assert False  # pytest tests/zen/test_zen_custom.py -k "test_custmer_node_cache_decision" -sv -l 


# @pytest.mark.asyncio
# async def test_aiohttp():
#     async with aiohttp.ClientSession() as session:
#         async with session.get('http://httpbin.org/get') as resp:
#             assert resp.status == 200
#             res = await resp.json()
#             assert res["url"] == 'http://httpbin.org/get'


# @pytest.mark.asyncio
# async def test_aiohttp2():
#     session = aiohttp.ClientSession()
#     async with session.get('http://httpbin.org/get') as resp:
#                 assert resp.status == 200
#                 res = await resp.json()
#                 assert res["url"] == 'http://httpbin.org/get'



from typing import Union
from fastapi import FastAPI, Request
app = FastAPI()


@app.get("/", tags=["test"]) 
@app.put("/", tags=["test"]) 
@app.post("/", tags=["test"]) 
def mytest(request: Request):
    return {"Hello": "World", "http_method": request.method}


@app.get("/items/{item_id}")
def read_item(item_id: int, q: Union[str, None] = None):
    return {"item_id": item_id, "q": q}











import logging
import random
import asyncio
from pprint import pprint
from pathlib import Path

from zen_rule import ZenRule, ast_exec

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


async def test_ast_exec():
    v = [{"name": "zoo", "args": [["'fccdjny'", "string"], [6, "int"], [3.14, "float"]], "ns": "", "id": "566a36f923b34cbd9c159272adc988ae"}, {"name": "bar", "args": [[{"name": "zoo", "args": [["'fccdjny'", "string"], [6, "int"], [3.14, "float"]], "ns": "", "id": "566a36f923b34cbd9c159272adc988ae"}, "func_value"], ["'a'", "string"]], "ns": "", "id": "5d0b81e086d748d4902a35fd85bad974"}, {"name": "bas", "args": [], "ns": "", "id": "7f8448aad2594b479f6df2e2241707c7"}, {"name": "foo", "args": [["myvar", "var"], [{"name": "bar", "args": [[{"name": "zoo", "args": [["'fccdjny'", "string"], [6, "int"], [3.14, "float"]], "ns": "", "id": "566a36f923b34cbd9c159272adc988ae"}, "func_value"], ["'a'", "string"]], "ns": "", "id": "5d0b81e086d748d4902a35fd85bad974"}, "func_value"], [{"name": "bas", "args": [], "ns": "", "id": "7f8448aad2594b479f6df2e2241707c7"}, "func_value"]], "ns": "", "id": "0c509d4654ef443eb621d791d3ffcaa1"}]

    await ast_exec(v, {"input": 7, "myvar": 15}, {"node_id": "nodexxxxx", "meta": {}})


async def test_zenrule_loader():
    def loader(key):
        print("key:", key)
        basedir = Path(__file__).parent
        filename = basedir / "graph" / key
        with open(filename, "r", encoding="utf8") as f:
            logger.warning(f"graph json: %s", filename)
            return f.read()

    # 测试同步和异步的 loader
    zr = ZenRule({"loader": loader})
    print(zr.engine.get_decision("custom_v2.json"))
    print(zr.engine.get_decision("custom_v2.json"))
    print(zr.engine.get_decision("custom_v2.json"))
    print(zr.engine.get_decision("custom_v2.json"))


async def test_zenengine_loader():
    import inspect
    from zen import ZenEngine
    def loader(key):
        print("key:", key)
        basedir = Path(__file__).parent
        filename = basedir / "graph" / key
        with open(filename, "r", encoding="utf8") as f:
            logger.warning(f"graph json: %s", filename)
            return f.read()

    # 测试同步和异步的 loader
    ze = ZenEngine({"loader": loader})
    print(inspect.iscoroutinefunction(ze.get_decision))
    # ze.get_decision("custom_v2.json") 如果 loader 是异步函数, 此函数调用会 hang 住. 
    # 这就是为什么不建议异步 loader 函数和 get_decision
    # print(inspect.iscoroutinefunction(ze.get_decision("custom_v2.json")))
    print(ze.get_decision("custom_v2.json"))
    print(ze.get_decision("custom_v2.json"))
    print(ze.get_decision("custom_v2.json"))
    print(ze.get_decision("custom_v2.json"))

    # result = await zr.async_evaluate("custom_v2_parse.json", {"input": 7, "myvar": 15})
    # print("zen rule custom_v2 result2:", result)


if __name__ == "__main__":
    asyncio.run(test_zenengine_loader())

    # import time
    # t1 = time.time()
    # asyncio.run(test_ast_exec())
    # t2 = time.time()
    # print("delta time:", t2 - t1)



import logging
import asyncio
from pprint import pprint
from pathlib import Path
from zen_rule.custom.udf_manager import udf, FuncArg, FuncRet
from zen_rule import ZenRule, ast_exec

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

@udf()
def zoo(*args, **kwargs):
    logger.info(f"  args:{args}")
    logger.info(f"kwargs:{kwargs}")
    return "zoo value"


@udf()
def bar(*args, **kwargs):
    logger.info(f"  args:{args}")
    logger.info(f"kwargs:{kwargs}")
    return "bar value"

@udf()
def bas(*args, **kwargs):
    logger.info(f"  args:{args}")
    logger.info(f"kwargs:{kwargs}")
    return "bas value"

@udf(
    comments="test udf foo",
    args_info=[
        FuncArg(arg_name="a", arg_type="string", defaults="", comments="var a"),
        FuncArg(arg_name="b", arg_type="string", defaults="", comments="var b"),
        FuncArg(arg_name="c", arg_type="string", defaults="", comments="var c"),
    ],
    return_info=FuncRet(field_type="string", examples="fccdjny", comments="返回值示例, 字段解释")     
)
def foo(*args, **kwargs):
    logger.info(f"  args:{args}")
    logger.info(f"kwargs:{kwargs}")
    return "foo value"


@udf(
    comments="group_distinct_1m_demo function",
    args_info=[
        FuncArg(arg_name="group", arg_type="string", defaults="", comments="var group"),
        FuncArg(arg_name="distinct", arg_type="string", defaults="", comments="var distinct"),
    ],
    return_info=FuncRet(field_type="string", examples={}, comments="返回值示例, 字段解释")     
)
def group_distinct_1m_demo(*args, **kwargs):
    logger.info(f"  args:{args}")
    logger.info(f"kwargs:{kwargs}")
    return {
            "function": "group_distinct_1m_demo",
            "pv": 1,
            "uv": 2,
            "gpv": 3
        }


async def test_zenrule():
    """
        推荐线上生产环境使用此模式进行规则执行, 可以缓存决策对象, 提高性能.
    """
    zr = ZenRule({})
    key = "xxxx_rule"
    basedir = Path(__file__).parent
    filename = basedir / "graph" / "custom_v2.json"

    with open(filename, "r", encoding="utf8") as f:
        logger.warning(f"graph json: %s", filename)
        content =  f.read()

    zr.create_decision_with_cache_key(key, content)  # 将规则图缓存在键下, 这样可以只读取规则一次，解析一次，然后复用决策对象 decision
    result = await zr.async_evaluate(key, {"input": 7, "myvar": 15})
    print("zen rule custom_v2 result:", result)

    result = await zr.async_evaluate(key, {"input": 7, "myvar": 15})
    print("zen rule custom_v2 result2:", result)



# 需要对 async loader 定义进行测试.
def loader(key):
    """
        loader 如果 loader 是异步函数, 那么同步的 get_decision 会有问题.
        除非我们自己实现 zenRule 的 get_decision, 而不是去调用 zenEngine的 get_decision
        加载规则还是让客户自己选择实现, 然后调用 create_decision_with_cache_key 缓存下来即可.
        暂时loader选择使用同步函数.
        此方法需要被覆写.
    """
    basedir = Path(__file__).parent
    filename = basedir / "graph" / key
    with open(filename, "r", encoding="utf8") as f:
        logger.warning(f"graph json: %s", filename)
        return f.read()


async def test_zenrule_with_loader():
    zr = ZenRule({"loader": loader})

    result = await zr.async_evaluate("custom_v2.json", {"input": 7, "myvar": 15})
    print("zen rule custom_v2 result:", result)

    result = await zr.async_evaluate("custom_v2.json", {"input": 7, "myvar": 15})
    print("zen rule custom_v2 result2:", result)

    result = await zr.async_evaluate("custom_v2_parse.json", {"input": 7, "myvar": 15})
    print("zen rule custom_v2 result2:", result)


if __name__ == "__main__":
    # test_zenrule
    asyncio.run(test_zenrule())
    asyncio.run(test_zenrule_with_loader())




import asyncio
import sys
from pprint import pprint, pformat
pprint(sys.path)
from custom.func_engine_v2 import zen_custom_expr_parse, ast_exec


if __name__ == "__main__":
    # "udf:foo(udf:bar(2  , rand(100),'a'), bas())"  "udf:foo(udf:bar(rand(100), udf:zoo(3,6),'a'), udf:bas())"
    # func_call_s = "udf:foo(myvar,udf:bar(rand(100), udf:zoo('fccdjny',6, 3.14),'a'), udf:bas())"
    func_call_s = "foo(myvar,bar(rand(100), zoo('fccdjny',6, 3.14),'a'), bas())"
    # func_call_s = "foo(bar( 2,'a'), bas())"

    expr_ast = zen_custom_expr_parse(func_call_s)
    print("--------------test case--------------------")
    print("expr    :", func_call_s)
    print("expr_ast:")
    pprint(expr_ast)
    import json
    rrr = json.dumps(expr_ast)
    print(rrr)

    res = asyncio.run(ast_exec(expr_ast, {"input": 7, "myvar": 15}))
    print("res:", res)